# -*- coding: utf-8 -*-
from django.conf import settings
from django.core.checks import Warning # pylint: disable=redefined-builtin
from django.utils.module_loading import import_string

from .utils import get_status_handler


def check_settings(**kwargs):
	issues = []

	status_handler = get_status_handler()

	if status_handler == 'django_email_log.status_handlers.celery.djcelery_email_handler':
		from .status_handlers.celery import DjceleryEmailTask
		config = getattr(settings, 'CELERY_EMAIL_TASK_CONFIG', None)

		if config is None:
			return [Warning(
				"Missing CELERY_EMAIL_TASK_CONFIG settings",
				hint="Add CELERY_EMAIL_TASK_CONFIG to settings",
				id='django_email_log.W001'
			)]
		if not 'base' in config:
			return [Warning(
				"Base attribute not in CELERY_EMAIL_TASK_CONFIG",
				hint="Add 'base': 'django_email_log.status_handlers.celery.DjceleryEmailTask' to CELERY_EMAIL_TASK_CONFIG",
				id='django_email_log.W002'
			)]
		base_name = str(config['base'])
		base = config['base']
		if isinstance(base, str):
			base = import_string(base)
		if not issubclass(base, DjceleryEmailTask):
			return [Warning(
				"%s is not subclass of django_email_log.status_handlers.celery.DjceleryEmailTask" % base_name,
				id='django_email_log.W003'
			)]
	return issues



# -*- coding: utf-8 -*-
import logging

from ..models import Email


logger = logging.getLogger(__name__)


def handler(status, email):
	try:
		sent = int(status or 0)
	except Exception:
		extra = {'status': status}
		if email is not None:
			extra.update({
				'email': email.original_message,
				'instance': email,
			})
		logger.exception("E-mail backend has returned invalid status", extra=extra)
		sent = 0
	if email is not None:
		email.status = Email.STATUS_SUCCESS if sent > 0 else Email.STATUS_FAIL
	return sent



# -*- coding: utf-8 -*-
from django.conf import settings


EMAIL_LOG_STATUS_HANDLER_MAP = getattr(settings, 'EMAIL_LOG_STATUS_HANDLER_MAP', {
	'': 'django_email_log.status_handlers.base.handler',
	'djcelery_email.backends.CeleryEmailBackend': 'django_email_log.status_handlers.celery.djcelery_email_handler'
})


def get_status_handler():
	handler = getattr(settings, 'EMAIL_LOG_STATUS_HANDLER', None)
	if handler is None:
		handler = EMAIL_LOG_STATUS_HANDLER_MAP.get(settings.EMAIL_LOG_BACKEND, EMAIL_LOG_STATUS_HANDLER_MAP[''])
	return handler



# -*- coding: utf-8 -*-
import logging
from contextvars import ContextVar as cv

from django.conf import settings
from django.core.mail import get_connection
from django.core.mail.backends.base import BaseEmailBackend
from django.utils import timezone
from django.utils.module_loading import import_string

from .models import Email
from .utils import get_status_handler


logger = logging.getLogger(__name__)
current_message = cv('current_email_message_instance')


class EmailBackend(BaseEmailBackend):
	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.connection = get_connection(settings.EMAIL_LOG_BACKEND, **kwargs)
		handler = get_status_handler()
		self.process_status = import_string(handler)

	def send_messages(self, email_messages):
		num_sent = 0
		for message in email_messages:
			message.connection = self.connection
			email = self.get_email_instance(message)
			current_message.set(email)
			try:
				num_sent += self.process_status(message.send(), email)
			except Exception:
				logger.exception("Wrong e-mail status value", extra={'value': num_sent})
			if email is not None:
				email.save()
		return num_sent

	def get_email_instance(self, message):
		if hasattr(message, 'model_instance'):
			model_instance = message.model_instance
		else:
			try:
				model_instance = Email.objects.create_from_message(message)
			except Exception:
				extra = {'email': message}
				logger.exception("E-mail message not serialized", extra=extra)
				model_instance = Email(
					subject='<error>',
					date_sent=timezone.now(),
					status=Email.STATUS_FAIL
				)
		if model_instance is not None:
			model_instance.original_message = message
		return model_instance



# -*- coding: utf-8 -*-
from django.conf import settings


EMAIL_LOG_BACKEND = getattr(settings, "EMAIL_LOG_BACKEND", 'django.core.mail.backends.console.EmailBackend')
EMAIL_LOG_STATUS_HANDLER = getattr(settings, "EMAIL_LOG_STATUS_HANDLER", None)
EMAIL_LOG_STATUS_HANDLER_MAP = getattr(settings, "EMAIL_LOG_STATUS_HANDLER_MAP", {
	'': 'email_log_bacend.status_handlers.base.handler',
})



# -*- coding: utf-8 -*-
from django.views.decorators.clickjacking import xframe_options_sameorigin
from django.urls import re_path

from .views import AttachmentView


urlpatterns = [
	re_path(r'^parts/(?P<object_type>(attachment|alternative))/(?P<pk>\d+)/(?:(?P<nr>\d+)/)?', xframe_options_sameorigin(AttachmentView.as_view()), name='django_email_log_attachment'),
]



# -*- coding: utf-8 -*-
import mimetypes
import os

from django.http.response import Http404, HttpResponse, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.views.generic import View

from .models import Email


class AttachmentView(View):
	def test_func(self):
		user = self.request.user
		return user.is_authenticated and user.is_staff and user.has_perm('django_email_log.change_email')

	def get(self, request, **kwargs):
		if not self.test_func():
			return HttpResponseForbidden()
		email = get_object_or_404(Email, pk=kwargs['pk'])
		payloads = email.payload_list
		number = int(kwargs.get('nr', '0'))
		if number > len(payloads):
			raise Http404()
		payload = payloads[number]
		part = payload.part
		content_type = part.get_content_type()
		if part.get_content_charset():
			content_type += '; charset=' + part.get_content_charset()

		if not 'nr' in kwargs:
			part_data = email.message_data
			content_type = 'text/plain; charset=utf-8'
		elif part.is_multipart():
			part_data = part.as_string()
			content_type = 'text/plain'
		else:
			part_data = part.get_payload(decode=True)
			charset = part.get_content_charset('iso-8859-1' if part.get_content_maintype() == 'text' else None)
			if charset == 'utf-8':
				try:
					part_data = part_data.decode('utf-8')
				except UnicodeDecodeError:
					part_data = part_data.decode('raw-unicode-escape')
			else:
				if charset:
					part_data = part_data.decode(charset, 'replace')
			if part.get_content_type() == 'text/html':
				for part in payloads:
					if part.cid:
						part_data = part_data.replace('cid:' + part.cid, part['get_absolute_url'])
		response = HttpResponse(part_data, content_type=content_type)
		if kwargs['object_type'] == 'attachment':
			ext = mimetypes.guess_extension(payload.content_type, strict=True)
			filename = payload.filename
			if filename is None:
				filename = email.date_sent.strftime('%Y-%m-%d_%H-%M-%S')
				filename = "%s_%d_%d" % (filename, email.pk, number)
			if ext:
				filename, __ = os.path.splitext(filename)
				filename = filename + ext
			response['Content-Disposition'] = 'attachment; filename="%s"' % filename.replace('"', '')
		return response



# -*- coding: utf-8 -*-
import logging

import celery

from ..backends import current_message
from ..models import Email


logger = logging.getLogger(__name__)


class DjceleryEmailTask(celery.Task):
	"""
	E-mail celery task instance with status capturing.
	"""

	def after_return(self, status, retval, task_id, args, kwargs, einfo):
		if status in ('SUCCESS', 'FAILURE', 'REVOKED') and 'email_id' in kwargs:
			try:
				success = (status == 'SUCCESS' and int(retval) > 0)
				email_status = Email.STATUS_SUCCESS if success else Email.STATUS_FAIL
			except Exception:
				email_status = Email.STATUS_FAIL
			Email.objects.filter(pk=kwargs['email_id']).update(status=email_status)

	def delay(self, *args, **kwargs):
		email_id = 0
		email = current_message.get()
		if email is not None:
			email_id = email.pk
		kwargs['email_id'] = email_id
		return super().delay(*args, **kwargs)


def djcelery_email_handler(status, email):
	"""
	Handles e-mail status if backend directly returns number of sent messages or
	do nothing if returns AsyncResult.
	"""
	try:
		sent = int(status or 0)
		email.status = Email.STATUS_SUCCESS if sent > 0 else Email.STATUS_FAIL
		return sent
	except Exception:
		return 0



# -*- coding: utf-8 -*-
from django.contrib import admin
from django.contrib.admin.utils import unquote
from django.utils.html import format_html
from django.utils.text import Truncator
from django.utils.translation import pgettext_lazy

from .models import Email


STATUS_NAMES = {
	Email.STATUS_PENDING: 'pending',
	Email.STATUS_SUCCESS: 'success',
	Email.STATUS_FAIL: 'fail',
	Email.STATUS_RECEIVED: 'received',
}


class EmailAdmin(admin.ModelAdmin):
	list_display = ('get_subject', 'get_recipients', 'date_sent')
	list_filter = ('date_sent', 'status', 'readed')
	search_fields = ('subject', 'email_to')
	readonly_fields = ('body', 'email_from', 'email_to', 'message_data', 'date_sent', 'status', 'readed')
	fields = ('email_from', 'email_to', 'date_sent', 'status')

	def get_recipients(self, obj):
		return Truncator(obj.email_to).chars(40, truncate="...")
	get_recipients.short_description = pgettext_lazy("email message", "recipients")
	get_recipients.allow_tags = True
	get_recipients.admin_order_field = "email_from"

	def has_add_permission(self, request): # pylint: disable=unused-argument
		return False

	def change_view(self, request, object_id, *args, **kwargs):
		message = self.get_object(request, unquote(object_id))
		if not message.readed:
			message.readed = True
			message.save(update_fields=['readed'])
		return super().change_view(request, object_id, *args, **kwargs)

	def get_subject(self, obj):
		status_name = STATUS_NAMES.get(obj.status, obj.status)
		status_icon = format_html('<span class="status {}"></span>', status_name)
		if obj.readed:
			return format_html('{}{}', status_icon, obj.subject)
		else:
			return format_html('<strong>{}{}</strong>', status_icon, obj.subject)
	get_subject.short_description = pgettext_lazy("email message", "subject")


admin.site.register(Email, EmailAdmin)



# Generated by Django 3.1.7 on 2021-03-21 10:45
import logging

from django.db import migrations, models, transaction
from django.db.models import Case, When, Q, Value as V, F
from django.db.models.functions import Cast, Replace
from django.db.utils import ProgrammingError


logger = logging.getLogger(__name__)


def transform_success_to_status(apps, schema_editor):
	db_alias = schema_editor.connection.alias
	Email = apps.get_model('django_email_log.Email')
	char_field = models.CharField(max_length=1)
	transform = Case(
		When(Q(success=True), then=V('s')),
		default=V('f'), output_field=char_field
	)
	Email.objects.using(db_alias).annotate(new_value=transform).update(status=F('new_value'))


def transform_status_to_success(apps, schema_editor):
	db_alias = schema_editor.connection.alias
	Email = apps.get_model('django_email_log.Email')
	bool_field = models.BooleanField()
	transform = Case(
		When(Q(status='f'), then=V(False)),
		default=V(True), output_field=bool_field
	)
	Email.objects.using(db_alias).annotate(new_value=transform).update(success=F('new_value'))


def copy_message_data(apps, schema_editor):
	db_alias = schema_editor.connection.alias
	Email = apps.get_model('django_email_log.Email')
	if schema_editor.connection.vendor == 'postgresql':
		with transaction.atomic():
			try:
				cursor = schema_editor.connection.cursor()
				cmd = 'ALTER TABLE "django_email_log_email" ALTER COLUMN "message_data_tmp" SET STORAGE EXTERNAL;'
				cursor.execute(cmd)
			except ProgrammingError as e:
				logger.warning(str(e))
				logger.warning("SQL command failed:\n%s\n" % cmd)
				transaction.rollback()
		Email.objects.using(db_alias).update(message_data_tmp=Cast(Replace('message_data', V('\\'), V('\\\\')), output_field=models.BinaryField()))
	else:
		Email.objects.using(db_alias).update(message_data_tmp=Cast('message_data', output_field=models.BinaryField()))


def uncopy_message_data(apps, schema_editor):
	db_alias = schema_editor.connection.alias
	Email = apps.get_model('django_email_log.Email')
	Email.objects.using(db_alias).update(message_data=Cast('message_data_tmp', output_field=models.TextField()))


class Migration(migrations.Migration):

	dependencies = [
		('django_email_log', '0001_initial'),
	]

	operations = [
		migrations.AddField(
			model_name='email',
			name='readed',
			field=models.BooleanField(blank=True, db_index=True, default=False, verbose_name='readed'),
		),
		migrations.AddField(
			model_name='email',
			name='status',
			field=models.CharField(choices=[('p', 'Pending'), ('s', 'Success'), ('f', 'Fail'), ('r', 'Received')], db_index=True, default='p', max_length=1, verbose_name='status'),
		),
		migrations.RunPython(transform_success_to_status, transform_status_to_success),
		migrations.RemoveField(
			model_name='email',
			name='success',
		),
		migrations.AlterField(
			model_name='email',
			name='message_data',
			field=models.TextField(verbose_name='data', default=''),
		),
		migrations.AddField(
			model_name='email',
			name='message_data_tmp',
			field=models.BinaryField(default=b''),
		),
		migrations.RunPython(copy_message_data, uncopy_message_data),
		migrations.RemoveField(
			model_name='email',
			name='message_data',
		),
		migrations.RenameField(
			model_name='email',
			old_name='message_data_tmp',
			new_name='message_data',
		),
		migrations.AlterField(
			model_name='email',
			name='message_data',
			field=models.BinaryField(),
		),
	]



# -*- coding: utf-8 -*-
from collections import namedtuple
import email

from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import pgettext_lazy


PartInfo = namedtuple('PartInfo', ['part', 'content_type', 'content_disposition', 'filename', 'cid', 'absolute_url', 'alternative_url'])


class EmailRawMessage(EmailMultiAlternatives):
	def __init__(self, raw, *args, **kwargs):
		self.__raw = raw
		super().__init__(*args, **kwargs)

	def _create_message(self, msg): # pylint: disable=unused-argument
		return self.__raw

	def message(self):
		return self._create_message(None)


class EmailManager(models.Manager):
	def create_from_message(self, message, commit=True):
		db_message = Email()
		db_message.subject = message.subject
		db_message.body = message.body
		msg = message.message()
		db_message.email_from = msg.get('From', '')
		db_message.email_to = msg.get('To', '')
		db_message.message_data = msg.as_bytes()
		if commit:
			db_message.save()
		return db_message


class Email(models.Model):
	objects = EmailManager()

	STATUS_PENDING = 'p'
	STATUS_SUCCESS = 's'
	STATUS_FAIL = 'f'
	STATUS_RECEIVED = 'r'
	STATUS_CHOICES = (
		(STATUS_PENDING, pgettext_lazy('email message status', "Pending")),
		(STATUS_SUCCESS, pgettext_lazy('email message status', "Success")),
		(STATUS_FAIL, pgettext_lazy('email message status', "Fail")),
		(STATUS_RECEIVED, pgettext_lazy('email message status', "Received")),
	)

	subject = models.TextField(verbose_name=pgettext_lazy('email message', "subject"))
	body = models.TextField(verbose_name=pgettext_lazy('email message', "body"))
	email_from = models.TextField(verbose_name=pgettext_lazy('email message', "sender"))
	email_to = models.TextField(verbose_name=pgettext_lazy('email message', "recipients"))
	message_data = models.BinaryField(verbose_name=pgettext_lazy('email message', "data"))

	date_sent = models.DateTimeField(verbose_name=pgettext_lazy('email message', "date sent"), editable=False, db_index=True)
	status = models.CharField(verbose_name=pgettext_lazy('email message', "status"), max_length=1, default=STATUS_PENDING, choices=STATUS_CHOICES, db_index=True)
	readed = models.BooleanField(verbose_name=pgettext_lazy('email message', "readed"), blank=True, default=False, db_index=True)

	def __str__(self):
		return self.subject

	class Meta:
		verbose_name = pgettext_lazy('email message noun', "e-mail message")
		verbose_name_plural = pgettext_lazy('email message noun plural', "e-mail messages")
		ordering = ('-date_sent',)

	def save(self, *args, **kwargs):
		if not self.id and not self.date_sent:
			self.date_sent = timezone.now()
		return super().save(*args, **kwargs)

	@property
	def parsed_message(self):
		return email.message_from_bytes(bytes(self.message_data))

	@property
	def payload_tree(self):
		return self.__preprocessed_message['tree']

	@property
	def payload_list(self):
		return self.__preprocessed_message['parts']

	@property
	def alternatives(self):
		return self.__preprocessed_message['alternatives']

	@property
	def __preprocessed_message(self):
		state = {'tree': {}, 'parts': [], 'alternatives': []}
		self.__build_tree(self.parsed_message, state['tree'], state, add_to_alternative=True)
		return state

	def __parseparam(self, s):
		while s[:1] == ';':
			s = s[1:]
			end = s.find(';')
			while end > 0 and (s.count('"', 0, end) - s.count('\\"', 0, end)) % 2:
				end = s.find(';', end + 1)
			if end < 0:
				end = len(s)
			f = s[:end]
			yield f.strip()
			s = s[end:]

	def __parse_content_disposition(self, line):
		"""
		From removed cgi module
		"""
		parts = self.__parseparam(';' + line)
		key = parts.__next__()
		pdict = {}
		for p in parts:
			i = p.find('=')
			if i >= 0:
				name = p[:i].strip().lower()
				value = p[i+1:].strip()
				if len(value) >= 2 and value[0] == value[-1] == '"':
					value = value[1:-1]
					value = value.replace('\\\\', '\\').replace('\\"', '"')
				pdict[name] = value
		return key, pdict

	def __build_tree(self, part, tree, state, add_to_alternative):
		parts, alternatives = state['parts'], state['alternatives']
		content_disposition, params = self.__parse_content_disposition(part.get('Content-Disposition', ''))
		filename = params.get('filename')
		content_disposition = content_disposition or 'inline'
		content_type = part.get_content_type()
		cid = part.get('Content-ID')
		if cid:
			if cid.startswith('<') and cid.endswith('>'):
				cid = cid[1:-1]
		tree['filename'] = filename
		tree['content_type'] = content_type
		tree['content_disposition'] = content_disposition
		tree['index'] = len(parts)
		tree['cid'] = len(parts)
		tree['absolute_url'] = reverse('django_email_log_attachment', args=('attachment', self.pk, tree['index']))
		tree['alternative_url'] = reverse('django_email_log_attachment', args=('alternative', self.pk, tree['index']))
		part_info = PartInfo(part, content_type, content_disposition, filename, cid, tree['absolute_url'], tree['alternative_url'])
		parts.append(part_info)
		if part.is_multipart():
			tree['children'] = []
			add_to_alternative = len(alternatives) == 0
			for payload in part.get_payload():
				subtree = {}
				tree['children'].append(subtree)
				self.__build_tree(payload, subtree, state, add_to_alternative)
		else:
			if add_to_alternative and part_info.content_disposition == 'inline':
				alternatives.append(part_info)
			tree['payload'] = part.get_payload()
		return part_info

	@property
	def email_message(self):
		msg = self.parsed_message
		email_instance = EmailRawMessage(
			raw=msg,
			from_email=msg.get('From'),
			to=msg['To'].split(', ') if 'To' in msg else None
		)
		setattr(email_instance, 'model_instance', self)
		return email_instance



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-03-20 14:08
from django.db import migrations, models


class Migration(migrations.Migration):

	initial = True

	dependencies = [
	]

	operations = [
		migrations.CreateModel(
			name='Email',
			fields=[
				('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
				('subject', models.TextField(verbose_name='subject')),
				('body', models.TextField(verbose_name='body')),
				('email_from', models.TextField(verbose_name='sender')),
				('email_to', models.TextField(verbose_name='recipients')),
				('message_data', models.TextField(verbose_name='message data')),
				('date_sent', models.DateTimeField(db_index=True, editable=False, verbose_name='date sent')),
				('success', models.BooleanField(db_index=True, default=False, verbose_name='successfully sent')),
			],
			options={
				'ordering': ('-date_sent',),
				'verbose_name': 'e-mail message',
				'verbose_name_plural': 'e-mail messages',
			},
		),
	]



# -*- coding: utf-8 -*-
from distutils.cmd import Command
from distutils.command.build import build as _build

import setuptools
from setuptools.command.install_lib import install_lib as _install_lib
from setuptools.command.develop import develop as _develop


class compile_translations(Command):
	description = 'Compiles translations"'
	user_options = []

	def initialize_options(self):
		pass

	def finalize_options(self):
		pass

	def run(self):
		import os

		if 'DJANGO_SETTINGS_MODULE' in os.environ:
			del os.environ['DJANGO_SETTINGS_MODULE']

		from django.core.management import call_command

		curdir = os.getcwd()
		call_command('compilemessages')
		os.chdir(curdir)


class build(_build):
	sub_commands = [('compile_translations', None)] + _build.sub_commands


class install_lib(_install_lib):
	def run(self):
		self.run_command('compile_translations')
		_install_lib.run(self)


class develop(_develop):
	def run(self):
		self.run_command('compile_translations')
		_develop.run(self)


with open('README.rst', 'r') as fh:
	long_description = fh.read()


setuptools.setup(
	name='django-email-log-backend',
	version='0.3.5',
	author='Miroslav Bendík',
	author_email='miroslav.bendik@gmail.com',
	description="Django email log backend",
	long_description=long_description,
	long_description_content_type='text/x-rst',
	url='https://github.com/mireq/django-email-log-backend',
	packages=setuptools.find_packages(),
	include_package_data=True,
	python_requires='>=3',
	setup_requires=['django'],
	cmdclass={'build': build, 'install_lib': install_lib, 'develop': develop, 'compile_translations': compile_translations},
	classifiers=[
		'Environment :: Web Environment',
		'Framework :: Django',
		'Intended Audience :: Developers',
		'Programming Language :: Python :: 3',
		'License :: OSI Approved :: MIT License',
		'Operating System :: OS Independent',
		'Topic :: Internet :: WWW/HTTP',
		'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
	],
)



# -*- coding: utf-8 -*-
from django.apps import AppConfig as CoreAppConfig
from django.core import checks
from django.utils.translation import gettext_lazy as _

from .checks import check_settings


class AppConfig(CoreAppConfig):
	name = 'django_email_log'
	verbose_name = _("Email log")

	def ready(self):
		checks.register()(check_settings)
